local modname = "yl_snowball"
local modpath = core.get_modpath(modname)

local Object = dofile(modpath .. DIR_DELIM .. "classic.lua")

local snowgame = {}


yl_snowball.GAME_TIMEOUT = 10 -- FIXME increase

local active_games = {}
local games_by_player = yl_snowball.games_by_player or {}
yl_snowball.games_by_player = games_by_player

local f = string.format

local game_hud = yl_snowball.game_hud

core.register_on_leaveplayer(
    function(player, _timed_out)
        local player_name = player:get_player_name()
        if player_name ~= "" then
            local game = games_by_player[player_name]
            if game then
                game:on_leaveplayer(player_name)
            end
            games_by_player[player_name] = nil
        end
    end
)

yl_snowball.update_hud = function(player)
    local player_name = player:get_player_name()
    local game = games_by_player[player_name]
    local text = "NOTHING"
    if game then
        text = game:get_info_string(true)
    else
        game_hud:set_enabled(player, false)
    end
    return {
        type = "text",
        text = text,
        number = 0xFFFFFF,
        direction = -1, -- right to left?
        position = { x = 0.7, y = 0.5 },
        alignment = { x = 1, y = -1 },
        offset = { x = -10, y = -10 },
        style = 1,
    }
end

if not game_hud then -- FIXME remove this
yl_snowball.game_hud = futil.define_hud( -- FIXME make sure not global
    "yl_snowball:hud_",
    {
        period = 0.567, -- FIXME not needed...?
        enabled_by_default = false,
        get_hud_def = function(player)
            return yl_snowball.update_hud(player)
        end,
    }
)
game_hud = yl_snowball.game_hud
end


local GamePlayer = Object:extend()
function GamePlayer:new(player_name)
    self.player_name = player_name
    self.active = true
    self.hits_taken = 0
    self.hits_given = 0
    self.shots = 0

    game_hud:set_enabled(core.get_player_by_name(player_name), true)
end


function GamePlayer:on_hit_take()
    self.hits_taken = self.hits_taken + 1
end


function GamePlayer:on_hit_give()
    self.hits_given = self.hits_given + 1

    -- this is a bit weird: since we don't start counting shots until
    -- first one hits, we may have more hits than shots :D
    self.shots = math.max(self.shots, self.hits_given)
end


function GamePlayer:on_shoot()
    self.shots = self.shots + 1
end

function GamePlayer:quit()
    --game_hud:set_enabled(core.get_player_by_name(self.player_name), false) -- this crashes XD
end


local Game = Object:extend()


function Game:new()
    local current_time = core.get_gametime()
    self.time_started = current_time
    self.time_last_action = current_time
    self.players = {}
    active_games[self] = true
    return self
end

function Game:get_time_left()
    local current_time = core.get_gametime()
    local time_left = yl_snowball.GAME_TIMEOUT - (current_time - self.time_last_action)
    return time_left > 0 and time_left or 0
end


function Game:get_info_string(show_time)
    local top = {}
    for p_name, player in pairs(self.players) do
        table.insert(top, {p_name, player})
    end
    table.sort(top, function (a,b) return a[2].hits_given > b[2].hits_given end)
    local lines = {}
    if show_time then
        table.insert(lines,  f("Game ends in: %ds", self:get_time_left()))
    end
    for _, info in ipairs(top) do
        local p_name = info[1]
        local player = info[2]
        local accuracy = player.hits_given/player.shots * 100
        table.insert(lines, f("%03d|%03d|%05.1f%%|%s",
                              player.hits_given,
                              player.hits_taken,
                              accuracy == accuracy and accuracy or 0,
                              p_name))
    end
    --print(dump(lines))
    return table.concat(lines, "\n")
end


function Game:add_player(player_name, game_player)
    -- local old_game = games_by_player[player_name]
    -- if not old_game then
    --     game_player = GamePlayer(player_name)
    -- else
    --     game_player = old_game:get_player(player_name)
    -- end
    if not game_player then
        game_player = GamePlayer(player_name)
    end
    self.players[player_name] = game_player
    games_by_player[player_name] = self
end


function Game:get_player(player_name)
    return self.players[player_name]
end


function Game:on_leaveplayer(player_name)
    game_hud:set_enabled(core.get_player_by_name(player_name), false)
    -- TODO set active = false
    games_by_player[player_name] = nil
end


function Game:merge(other_game)
    for name, game_player in pairs(other_game.players) do
        self:add_player(name, game_player)
    end
    self.time_started = math.min(self.time_started, other_game.time_started)
    self.time_last_action = math.max(self.time_last_action, other_game.time_last_action)
    active_games[other_game] = nil
end


function Game:register_hit(shooter_name, target_name)
    local s_player = self:get_player(shooter_name)
    s_player:on_hit_give()

    local t_player = self:get_player(target_name)
    t_player:on_hit_take()
    -- print(shooter_name, dump(s_player))
    -- print(target_name, dump(t_player))

    --local text = self:get_info_string()
    --print(text)

    local current_time = core.get_gametime()
    self.time_last_action = current_time
end


function Game:register_shot(shooter_name)
    local s_player = self:get_player(shooter_name)
    s_player:on_shoot()
    local current_time = core.get_gametime()
    self.time_last_action = current_time
end


function Game:stop()
    for p_name, player in pairs(self.players) do
        -- TODO show only top + your own scores
        local lasted = self.time_last_action - self.time_started
        local lines = {
            f("* Snowball Game Ended (lasted %ds)", lasted),
            self:get_info_string(),
        }
        core.chat_send_player(p_name, table.concat(lines, "\n"))
    end

    for player_name, player in pairs(self.players) do
        player:quit()
        games_by_player[player_name] = nil
    end
    active_games[self] = nil
end


function snowgame.register_hit(shooter_name, target_name)
    -- TODO check here and bail if target is opted-out from snowball games!

    -- print(shooter_name, target_name)
    local s_game = games_by_player[shooter_name]
    local t_game = games_by_player[target_name]

    if s_game and t_game and (s_game ~= t_game) then
        -- if both player but different games, merge
        s_game:merge(t_game)
    end
    local game = s_game or t_game
    if not game then
        -- none are playing, make a new one
        game = Game()
    end
    if not s_game then
        -- add shooter to current
        game:add_player(shooter_name)
        game:register_shot(shooter_name)
    end
    if not t_game then
        -- add target to current (if played diff game, was merged)
        game:add_player(target_name)
    end
    game:register_hit(shooter_name, target_name)
end


function snowgame.register_shot(shooter_name)
    local s_game = games_by_player[shooter_name]
    if s_game then
        s_game:register_shot(shooter_name)
    end
end


futil.register_globalstep({
    period = 1.31,
    func = function(dtime)
        local current_time = core.get_gametime()
        for game,_ in pairs(active_games) do
            if (current_time - game.time_last_action) > yl_snowball.GAME_TIMEOUT then
                game:stop()
            end
        end
    end,
})


yl_snowball.snowgame = snowgame

return snowgame
